library angular2.src.core.compiler.compiler;

import 'package:angular2/di.dart' show Injectable;
import 'package:angular2/src/facade/lang.dart'
    show Type, isBlank, isPresent, BaseException, normalizeBlank, stringify;
import 'package:angular2/src/facade/async.dart' show Future, PromiseWrapper;
import 'package:angular2/src/facade/collection.dart'
    show List, ListWrapper, Map, MapWrapper;
import 'package:angular2/change_detection.dart' show ChangeDetection, Parser;
import './directive_metadata_reader.dart' show DirectiveMetadataReader;
import './view.dart' show ProtoView;
import './pipeline/compile_pipeline.dart' show CompilePipeline;
import './pipeline/compile_element.dart' show CompileElement;
import './pipeline/default_steps.dart' show createDefaultSteps;
import './template_loader.dart' show TemplateLoader;
import './template_resolver.dart' show TemplateResolver;
import '../annotations/template.dart' show Template;
import './shadow_dom_strategy.dart' show ShadowDomStrategy;
import './pipeline/compile_step.dart' show CompileStep;
import './component_url_mapper.dart' show ComponentUrlMapper;
import './url_resolver.dart' show UrlResolver;
import './css_processor.dart' show CssProcessor;

@Injectable()
class CompilerCache {
  Map _cache;
  CompilerCache() {
    this._cache = MapWrapper.create();
  }
  set(Type component, ProtoView protoView) {
    MapWrapper.set(this._cache, component, protoView);
  }
  ProtoView get(Type component) {
    var result = MapWrapper.get(this._cache, component);
    return normalizeBlank(result);
  }
  clear() {
    MapWrapper.clear(this._cache);
  }
}
@Injectable()
class Compiler {
  DirectiveMetadataReader _reader;
  Parser _parser;
  CompilerCache _compilerCache;
  ChangeDetection _changeDetection;
  TemplateLoader _templateLoader;
  Map<Type, Future> _compiling;
  ShadowDomStrategy _shadowDomStrategy;
  TemplateResolver _templateResolver;
  ComponentUrlMapper _componentUrlMapper;
  UrlResolver _urlResolver;
  String _appUrl;
  CssProcessor _cssProcessor;
  Compiler(ChangeDetection changeDetection, TemplateLoader templateLoader,
      DirectiveMetadataReader reader, Parser parser, CompilerCache cache,
      ShadowDomStrategy shadowDomStrategy, TemplateResolver templateResolver,
      ComponentUrlMapper componentUrlMapper, UrlResolver urlResolver,
      CssProcessor cssProcessor) {
    this._changeDetection = changeDetection;
    this._reader = reader;
    this._parser = parser;
    this._compilerCache = cache;
    this._templateLoader = templateLoader;
    this._compiling = MapWrapper.create();
    this._shadowDomStrategy = shadowDomStrategy;
    this._templateResolver = templateResolver;
    this._componentUrlMapper = componentUrlMapper;
    this._urlResolver = urlResolver;
    this._appUrl = urlResolver.resolve(null, './');
    this._cssProcessor = cssProcessor;
  }
  List<CompileStep> createSteps(Type component, Template template) {
    var dirMetadata = ListWrapper.map(this._flattenDirectives(template), (d) {
      return this._reader.read(d);
    });
    var cmpMetadata = this._reader.read(component);
    var templateUrl = this._templateLoader.getTemplateUrl(template);
    return createDefaultSteps(this._changeDetection, this._parser, cmpMetadata,
        dirMetadata, this._shadowDomStrategy, templateUrl, this._cssProcessor);
  }
  Future<ProtoView> compile(Type component) {
    var protoView = this._compile(component);
    return PromiseWrapper.isPromise(protoView)
        ? protoView
        : PromiseWrapper.resolve(protoView);
  }
  _compile(Type component) {
    var protoView = this._compilerCache.get(component);
    if (isPresent(protoView)) {
      return protoView;
    }
    var pvPromise = MapWrapper.get(this._compiling, component);
    if (isPresent(pvPromise)) {
      return pvPromise;
    }
    var template = this._templateResolver.resolve(component);
    var componentUrl = this._componentUrlMapper.getUrl(component);
    var baseUrl = this._urlResolver.resolve(this._appUrl, componentUrl);
    this._templateLoader.setBaseUrl(template, baseUrl);
    var tplElement = this._templateLoader.load(template);
    if (PromiseWrapper.isPromise(tplElement)) {
      pvPromise = PromiseWrapper.then(tplElement, (el) {
        return this._compileTemplate(template, el, component);
      }, (_) {
        throw new BaseException(
            '''Failed to load the template for ${stringify(component)}''');
      });
      MapWrapper.set(this._compiling, component, pvPromise);
      return pvPromise;
    }
    return this._compileTemplate(template, tplElement, component);
  }
  _compileTemplate(Template template, tplElement, Type component) {
    var pipeline = new CompilePipeline(this.createSteps(component, template));
    var compileElements;
    try {
      compileElements = pipeline.process(tplElement, stringify(component));
    } catch (ex) {
      return PromiseWrapper.reject(ex);
    }
    var protoView = compileElements[0].inheritedProtoView;
    this._compilerCache.set(component, protoView);
    MapWrapper.delete(this._compiling, component);
    var nestedPVPromises = [];
    for (var i = 0; i < compileElements.length; i++) {
      var ce = compileElements[i];
      if (ce.hasNestedView) {
        this._compileNestedProtoView(ce, nestedPVPromises);
      }
    }
    if (protoView.stylePromises.length > 0) {
      var syncProtoView = protoView;
      protoView = PromiseWrapper.all(syncProtoView.stylePromises).then((_) {
        return syncProtoView;
      });
    }
    if (nestedPVPromises.length > 0) {
      return PromiseWrapper.then(PromiseWrapper.all(nestedPVPromises), (_) {
        return protoView;
      }, (e) {
        throw new BaseException(
            '''${e.message} -> Failed to compile ${stringify(component)}''');
      });
    }
    return protoView;
  }
  _compileNestedProtoView(CompileElement ce, List<Future> promises) {
    var protoView = this._compile(ce.componentDirective.type);
    if (PromiseWrapper.isPromise(protoView)) {
      ListWrapper.push(promises, protoView.then((pv) {
        ce.inheritedElementBinder.nestedProtoView = pv;
      }));
    } else {
      ce.inheritedElementBinder.nestedProtoView = protoView;
    }
  }
  List<Type> _flattenDirectives(Template template) {
    if (isBlank(template.directives)) return [];
    var directives = [];
    this._flattenList(template.directives, directives);
    return directives;
  }
  _flattenList(List<dynamic> tree, List<Type> out) {
    for (var i = 0; i < tree.length; i++) {
      var item = tree[i];
      if (ListWrapper.isList(item)) {
        this._flattenList(item, out);
      } else {
        ListWrapper.push(out, item);
      }
    }
  }
}
