library angular2.src.core.compiler.pipeline.element_binder_builder;

import 'package:angular2/src/facade/lang.dart'
    show
        int,
        isPresent,
        isBlank,
        Type,
        BaseException,
        StringWrapper,
        RegExpWrapper,
        isString,
        stringify;
import 'package:angular2/src/dom/dom_adapter.dart' show DOM;
import 'package:angular2/src/facade/collection.dart'
    show ListWrapper, List, MapWrapper, StringMapWrapper;
import 'package:angular2/src/reflection/reflection.dart' show reflector;
import 'package:angular2/change_detection.dart'
    show Parser, ProtoChangeDetector;
import '../directive_metadata.dart' show DirectiveMetadata;
import './compile_step.dart' show CompileStep;
import './compile_element.dart' show CompileElement;
import './compile_control.dart' show CompileControl;
import './util.dart' show dashCaseToCamelCase, camelCaseToDashCase;

var DOT_REGEXP = RegExpWrapper.create('\\.');
const ATTRIBUTE_PREFIX = 'attr.';
var attributeSettersCache = StringMapWrapper.create();
_isValidAttributeValue(String attrName, dynamic value) {
  if (attrName == "role") {
    return isString(value);
  } else {
    return isPresent(value);
  }
}
attributeSetterFactory(String attrName) {
  var setterFn = StringMapWrapper.get(attributeSettersCache, attrName);
  var dashCasedAttributeName;
  if (isBlank(setterFn)) {
    dashCasedAttributeName = camelCaseToDashCase(attrName);
    setterFn = (element, value) {
      if (_isValidAttributeValue(dashCasedAttributeName, value)) {
        DOM.setAttribute(element, dashCasedAttributeName, stringify(value));
      } else {
        DOM.removeAttribute(element, dashCasedAttributeName);
        if (isPresent(value)) {
          throw new BaseException("Invalid " +
              dashCasedAttributeName +
              " attribute, only string values are allowed, got '" +
              stringify(value) +
              "'");
        }
      }
    };
    StringMapWrapper.set(attributeSettersCache, attrName, setterFn);
  }
  return setterFn;
}
const CLASS_PREFIX = 'class.';
var classSettersCache = StringMapWrapper.create();
classSetterFactory(String className) {
  var setterFn = StringMapWrapper.get(classSettersCache, className);
  if (isBlank(setterFn)) {
    setterFn = (element, value) {
      if (value) {
        DOM.addClass(element, className);
      } else {
        DOM.removeClass(element, className);
      }
    };
    StringMapWrapper.set(classSettersCache, className, setterFn);
  }
  return setterFn;
}
const STYLE_PREFIX = 'style.';
var styleSettersCache = StringMapWrapper.create();
styleSetterFactory(String styleName, String stylesuffix) {
  var cacheKey = styleName + stylesuffix;
  var setterFn = StringMapWrapper.get(styleSettersCache, cacheKey);
  var dashCasedStyleName;
  if (isBlank(setterFn)) {
    dashCasedStyleName = camelCaseToDashCase(styleName);
    setterFn = (element, value) {
      var valAsStr;
      if (isPresent(value)) {
        valAsStr = stringify(value);
        DOM.setStyle(element, dashCasedStyleName, valAsStr + stylesuffix);
      } else {
        DOM.removeStyle(element, dashCasedStyleName);
      }
    };
    StringMapWrapper.set(classSettersCache, cacheKey, setterFn);
  }
  return setterFn;
}
class ElementBinderBuilder extends CompileStep {
  Parser _parser;
  ElementBinderBuilder(Parser parser) : super() {
    this._parser = parser;
  }
  process(
      CompileElement parent, CompileElement current, CompileControl control) {
    var elementBinder = null;
    var parentElementBinder = null;
    var distanceToParentBinder =
        this._getDistanceToParentBinder(parent, current);
    if (isPresent(parent)) {
      parentElementBinder = parent.inheritedElementBinder;
    }
    if (current.hasBindings) {
      var protoView = current.inheritedProtoView;
      var protoInjectorWasBuilt = isBlank(parent)
          ? true
          : !identical(current.inheritedProtoElementInjector,
              parent.inheritedProtoElementInjector);
      var currentProtoElementInjector =
          protoInjectorWasBuilt ? current.inheritedProtoElementInjector : null;
      elementBinder = protoView.bindElement(parentElementBinder,
          distanceToParentBinder, currentProtoElementInjector,
          current.componentDirective, current.viewportDirective);
      current.distanceToParentBinder = 0;
      if (isPresent(current.textNodeBindings)) {
        this._bindTextNodes(protoView, current);
      }
      if (isPresent(current.propertyBindings)) {
        this._bindElementProperties(protoView, current);
      }
      if (isPresent(current.eventBindings)) {
        this._bindEvents(protoView, current);
      }
      if (isPresent(current.contentTagSelector)) {
        elementBinder.contentTagSelector = current.contentTagSelector;
      }
      var directives = current.getAllDirectives();
      this._bindDirectiveProperties(directives, current);
      this._bindDirectiveEvents(directives, current);
    } else if (isPresent(parent)) {
      elementBinder = parentElementBinder;
      current.distanceToParentBinder = distanceToParentBinder;
    }
    current.inheritedElementBinder = elementBinder;
  }
  _getDistanceToParentBinder(parent, current) {
    return isPresent(parent) ? parent.distanceToParentBinder + 1 : 0;
  }
  _bindTextNodes(protoView, compileElement) {
    MapWrapper.forEach(compileElement.textNodeBindings,
        (expression, indexInParent) {
      protoView.bindTextNode(indexInParent, expression);
    });
  }
  _bindElementProperties(protoView, compileElement) {
    MapWrapper.forEach(compileElement.propertyBindings, (expression, property) {
      var setterFn;
      var styleParts;
      var styleSuffix;
      if (StringWrapper.startsWith(property, ATTRIBUTE_PREFIX)) {
        setterFn = attributeSetterFactory(
            StringWrapper.substring(property, ATTRIBUTE_PREFIX.length));
      } else if (StringWrapper.startsWith(property, CLASS_PREFIX)) {
        setterFn = classSetterFactory(
            StringWrapper.substring(property, CLASS_PREFIX.length));
      } else if (StringWrapper.startsWith(property, STYLE_PREFIX)) {
        styleParts = StringWrapper.split(property, DOT_REGEXP);
        styleSuffix =
            styleParts.length > 2 ? ListWrapper.get(styleParts, 2) : '';
        setterFn =
            styleSetterFactory(ListWrapper.get(styleParts, 1), styleSuffix);
      } else if (StringWrapper.equals(property, 'innerHtml')) {
        setterFn = (element, value) {
          return DOM.setInnerHTML(element, value);
        };
      } else {
        property = this._resolvePropertyName(property);
        var propertySetterFn = reflector.setter(property);
        setterFn = (receiver, value) {
          if (DOM.hasProperty(receiver, property)) {
            return propertySetterFn(receiver, value);
          }
        };
      }
      protoView.bindElementProperty(expression.ast, property, setterFn);
    });
  }
  _bindEvents(protoView, compileElement) {
    MapWrapper.forEach(compileElement.eventBindings, (expression, eventName) {
      protoView.bindEvent(eventName, expression);
    });
  }
  _bindDirectiveEvents(
      List<DirectiveMetadata> directives, CompileElement compileElement) {
    for (var directiveIndex = 0;
        directiveIndex < directives.length;
        directiveIndex++) {
      var directive = directives[directiveIndex];
      var annotation = directive.annotation;
      if (isBlank(annotation.events)) continue;
      var protoView = compileElement.inheritedProtoView;
      StringMapWrapper.forEach(annotation.events, (action, eventName) {
        var expression =
            this._parser.parseAction(action, compileElement.elementDescription);
        protoView.bindEvent(eventName, expression, directiveIndex);
      });
    }
  }
  _bindDirectiveProperties(
      List<DirectiveMetadata> directives, CompileElement compileElement) {
    var protoView = compileElement.inheritedProtoView;
    for (var directiveIndex = 0;
        directiveIndex < directives.length;
        directiveIndex++) {
      var directive = ListWrapper.get(directives, directiveIndex);
      var annotation = directive.annotation;
      if (isBlank(annotation.bind)) continue;
      StringMapWrapper.forEach(annotation.bind, (bindConfig, dirProp) {
        var pipes = this._splitBindConfig(bindConfig);
        var elProp = ListWrapper.removeAt(pipes, 0);
        var bindingAst = isPresent(compileElement.propertyBindings)
            ? MapWrapper.get(
                compileElement.propertyBindings, dashCaseToCamelCase(elProp))
            : null;
        if (isBlank(bindingAst)) {
          var attributeValue = MapWrapper.get(compileElement.attrs(), elProp);
          if (isPresent(attributeValue)) {
            bindingAst = this._parser.wrapLiteralPrimitive(
                attributeValue, compileElement.elementDescription);
          }
        }
        if (isPresent(bindingAst)) {
          var fullExpAstWithBindPipes =
              this._parser.addPipes(bindingAst, pipes);
          protoView.bindDirectiveProperty(directiveIndex,
              fullExpAstWithBindPipes, dirProp,
              reflector.setter(dashCaseToCamelCase(dirProp)));
        }
      });
    }
  }
  _splitBindConfig(String bindConfig) {
    return ListWrapper.map(bindConfig.split('|'), (s) {
      return s.trim();
    });
  }
  _resolvePropertyName(String attrName) {
    var mappedPropName = StringMapWrapper.get(DOM.attrToPropMap, attrName);
    return isPresent(mappedPropName) ? mappedPropName : attrName;
  }
}
