library angular2.src.di.binding;

import 'package:angular2/src/facade/lang.dart' show Type, isBlank, isPresent;
import 'package:angular2/src/facade/collection.dart'
    show List, MapWrapper, ListWrapper;
import 'package:angular2/src/reflection/reflection.dart' show reflector;
import './key.dart' show Key;
import './annotations.dart'
    show Inject, InjectLazy, InjectPromise, Optional, DependencyAnnotation;
import './exceptions.dart' show NoAnnotationError;

class Dependency {
  Key key;
  bool asPromise;
  bool lazy;
  bool optional;
  List properties;
  Dependency(
      Key key, bool asPromise, bool lazy, bool optional, List properties) {
    this.key = key;
    this.asPromise = asPromise;
    this.lazy = lazy;
    this.optional = optional;
    this.properties = properties;
  }
  static fromKey(Key key) {
    return new Dependency(key, false, false, false, []);
  }
}
class Binding {
  Key key;
  Function factory;
  List dependencies;
  bool providedAsPromise;
  Binding(
      Key key, Function factory, List dependencies, bool providedAsPromise) {
    this.key = key;
    this.factory = factory;
    this.dependencies = dependencies;
    this.providedAsPromise = providedAsPromise;
  }
}
BindingBuilder bind(token) {
  return new BindingBuilder(token);
}
class BindingBuilder {
  var token;
  BindingBuilder(token) {
    this.token = token;
  }
  Binding toClass(Type type) {
    return new Binding(Key.get(this.token), reflector.factory(type),
        _dependenciesFor(type), false);
  }
  Binding toValue(value) {
    return new Binding(Key.get(this.token), () {
      return value;
    }, [], false);
  }
  Binding toAlias(aliasToken) {
    return new Binding(Key.get(this.token), (aliasInstance) {
      return aliasInstance;
    }, [Dependency.fromKey(Key.get(aliasToken))], false);
  }
  Binding toFactory(Function factoryFunction, [List dependencies = null]) {
    return new Binding(Key.get(this.token), factoryFunction,
        this._constructDependencies(factoryFunction, dependencies), false);
  }
  Binding toAsyncFactory(Function factoryFunction, [List dependencies = null]) {
    return new Binding(Key.get(this.token), factoryFunction,
        this._constructDependencies(factoryFunction, dependencies), true);
  }
  _constructDependencies(Function factoryFunction, List dependencies) {
    return isBlank(dependencies)
        ? _dependenciesFor(factoryFunction)
        : ListWrapper.map(dependencies, (t) {
      return Dependency.fromKey(Key.get(t));
    });
  }
}
List _dependenciesFor(typeOrFunc) {
  var params = reflector.parameters(typeOrFunc);
  if (isBlank(params)) return [];
  if (ListWrapper.any(params, (p) {
    return isBlank(p);
  })) throw new NoAnnotationError(typeOrFunc);
  return ListWrapper.map(params, (p) {
    return _extractToken(typeOrFunc, p);
  });
}
_extractToken(typeOrFunc, annotations) {
  var depProps = [];
  var token = null;
  var optional = false;
  var lazy = false;
  var asPromise = false;
  for (var i = 0; i < annotations.length; ++i) {
    var paramAnnotation = annotations[i];
    if (paramAnnotation is Type) {
      token = paramAnnotation;
    } else if (paramAnnotation is Inject) {
      token = paramAnnotation.token;
    } else if (paramAnnotation is InjectPromise) {
      token = paramAnnotation.token;
      asPromise = true;
    } else if (paramAnnotation is InjectLazy) {
      token = paramAnnotation.token;
      lazy = true;
    } else if (paramAnnotation is Optional) {
      optional = true;
    } else if (paramAnnotation is DependencyAnnotation) {
      ListWrapper.push(depProps, paramAnnotation);
    }
  }
  if (isPresent(token)) {
    return _createDependency(token, asPromise, lazy, optional, depProps);
  } else {
    throw new NoAnnotationError(typeOrFunc);
  }
}
Dependency _createDependency(token, asPromise, lazy, optional, depProps) {
  return new Dependency(Key.get(token), asPromise, lazy, optional, depProps);
}
