library angular2.src.di.injector;

import 'package:angular2/src/facade/collection.dart'
    show Map, List, MapWrapper, ListWrapper;
import './binding.dart' show Binding, BindingBuilder, bind;
import './exceptions.dart'
    show
        ProviderError,
        NoProviderError,
        InvalidBindingError,
        AsyncBindingError,
        CyclicDependencyError,
        InstantiationError;
import 'package:angular2/src/facade/lang.dart'
    show FunctionWrapper, Type, isPresent, isBlank;
import 'package:angular2/src/facade/async.dart' show Future, PromiseWrapper;
import './key.dart' show Key;

var _constructing = new Object();
var _notFound = new Object();
class _Waiting {
  Future promise;
  _Waiting(Future promise) {
    this.promise = promise;
  }
}
bool _isWaiting(obj) {
  return obj is _Waiting;
}
class Injector {
  List _bindings;
  List _instances;
  Injector _parent;
  bool _defaultBindings;
  _AsyncInjectorStrategy _asyncStrategy;
  _SyncInjectorStrategy _syncStrategy;
  Injector(List bindings, {parent: null, defaultBindings: false}) {
    var flatten = _flattenBindings(bindings, MapWrapper.create());
    this._bindings = this._createListOfBindings(flatten);
    this._instances = this._createInstances();
    this._parent = parent;
    this._defaultBindings = defaultBindings;
    this._asyncStrategy = new _AsyncInjectorStrategy(this);
    this._syncStrategy = new _SyncInjectorStrategy(this);
  }
  get(token) {
    return this._getByKey(Key.get(token), false, false, false);
  }
  getOptional(token) {
    return this._getByKey(Key.get(token), false, false, true);
  }
  asyncGet(token) {
    return this._getByKey(Key.get(token), true, false, false);
  }
  Injector createChild(List bindings) {
    return new Injector(bindings, parent: this);
  }
  List _createListOfBindings(flattenBindings) {
    var bindings = ListWrapper.createFixedSize(Key.numberOfKeys + 1);
    MapWrapper.forEach(flattenBindings, (v, keyId) {
      return bindings[keyId] = v;
    });
    return bindings;
  }
  List _createInstances() {
    return ListWrapper.createFixedSize(Key.numberOfKeys + 1);
  }
  _getByKey(Key key, bool returnPromise, bool returnLazy, bool optional) {
    if (returnLazy) {
      return () {
        return this._getByKey(key, returnPromise, false, optional);
      };
    }
    var strategy = returnPromise ? this._asyncStrategy : this._syncStrategy;
    var instance = strategy.readFromCache(key);
    if (!identical(instance, _notFound)) return instance;
    instance = strategy.instantiate(key);
    if (!identical(instance, _notFound)) return instance;
    if (isPresent(this._parent)) {
      return this._parent._getByKey(key, returnPromise, returnLazy, optional);
    }
    if (optional) {
      return null;
    } else {
      throw new NoProviderError(key);
    }
  }
  List _resolveDependencies(Key key, Binding binding, bool forceAsync) {
    try {
      var getDependency = (d) {
        return this._getByKey(
            d.key, forceAsync || d.asPromise, d.lazy, d.optional);
      };
      return ListWrapper.map(binding.dependencies, getDependency);
    } catch (e) {
      this._clear(key);
      if (e is ProviderError) e.addKey(key);
      throw e;
    }
  }
  _getInstance(Key key) {
    if (this._instances.length <= key.id) return null;
    return ListWrapper.get(this._instances, key.id);
  }
  _setInstance(Key key, obj) {
    ListWrapper.set(this._instances, key.id, obj);
  }
  _getBinding(Key key) {
    var binding = this._bindings.length <= key.id
        ? null
        : ListWrapper.get(this._bindings, key.id);
    if (isBlank(binding) && this._defaultBindings) {
      return bind(key.token).toClass(key.token);
    } else {
      return binding;
    }
  }
  _markAsConstructing(Key key) {
    this._setInstance(key, _constructing);
  }
  _clear(Key key) {
    this._setInstance(key, null);
  }
}
class _SyncInjectorStrategy {
  Injector injector;
  _SyncInjectorStrategy(Injector injector) {
    this.injector = injector;
  }
  readFromCache(Key key) {
    if (identical(key.token, Injector)) {
      return this.injector;
    }
    var instance = this.injector._getInstance(key);
    if (identical(instance, _constructing)) {
      throw new CyclicDependencyError(key);
    } else if (isPresent(instance) && !_isWaiting(instance)) {
      return instance;
    } else {
      return _notFound;
    }
  }
  instantiate(Key key) {
    var binding = this.injector._getBinding(key);
    if (isBlank(binding)) return _notFound;
    if (binding.providedAsPromise) throw new AsyncBindingError(key);
    this.injector._markAsConstructing(key);
    var deps = this.injector._resolveDependencies(key, binding, false);
    return this._createInstance(key, binding, deps);
  }
  _createInstance(Key key, Binding binding, List deps) {
    try {
      var instance = FunctionWrapper.apply(binding.factory, deps);
      this.injector._setInstance(key, instance);
      return instance;
    } catch (e) {
      this.injector._clear(key);
      throw new InstantiationError(e, key);
    }
  }
}
class _AsyncInjectorStrategy {
  Injector injector;
  _AsyncInjectorStrategy(Injector injector) {
    this.injector = injector;
  }
  readFromCache(Key key) {
    if (identical(key.token, Injector)) {
      return PromiseWrapper.resolve(this.injector);
    }
    var instance = this.injector._getInstance(key);
    if (identical(instance, _constructing)) {
      throw new CyclicDependencyError(key);
    } else if (_isWaiting(instance)) {
      return instance.promise;
    } else if (isPresent(instance)) {
      return PromiseWrapper.resolve(instance);
    } else {
      return _notFound;
    }
  }
  instantiate(Key key) {
    var binding = this.injector._getBinding(key);
    if (isBlank(binding)) return _notFound;
    this.injector._markAsConstructing(key);
    var deps = this.injector._resolveDependencies(key, binding, true);
    var depsPromise = PromiseWrapper.all(deps);
    var promise = PromiseWrapper.then(depsPromise, null, (e) {
      return this._errorHandler(key, e);
    }).then((deps) {
      return this._findOrCreate(key, binding, deps);
    }).then((instance) {
      return this._cacheInstance(key, instance);
    });
    this.injector._setInstance(key, new _Waiting(promise));
    return promise;
  }
  Future _errorHandler(Key key, e) {
    if (e is ProviderError) e.addKey(key);
    return PromiseWrapper.reject(e);
  }
  _findOrCreate(Key key, Binding binding, List deps) {
    try {
      var instance = this.injector._getInstance(key);
      if (!_isWaiting(instance)) return instance;
      return FunctionWrapper.apply(binding.factory, deps);
    } catch (e) {
      this.injector._clear(key);
      throw new InstantiationError(e, key);
    }
  }
  _cacheInstance(key, instance) {
    this.injector._setInstance(key, instance);
    return instance;
  }
}
_flattenBindings(List bindings, Map res) {
  ListWrapper.forEach(bindings, (b) {
    if (b is Binding) {
      MapWrapper.set(res, b.key.id, b);
    } else if (b is Type) {
      var s = bind(b).toClass(b);
      MapWrapper.set(res, s.key.id, s);
    } else if (b is List) {
      _flattenBindings(b, res);
    } else if (b is BindingBuilder) {
      throw new InvalidBindingError(b.token);
    } else {
      throw new InvalidBindingError(b);
    }
  });
  return res;
}
