library angular2.test.core.compiler.shadow_dom.light_dom_spec;

import 'package:angular2/test_lib.dart'
    show describe, beforeEach, it, expect, ddescribe, iit, SpyObject, el, proxy;
import 'package:angular2/src/facade/lang.dart'
    show IMPLEMENTS, isBlank, isPresent;
import 'package:angular2/src/facade/collection.dart'
    show ListWrapper, MapWrapper;
import 'package:angular2/src/dom/dom_adapter.dart' show DOM;
import 'package:angular2/src/core/compiler/shadow_dom_emulation/content_tag.dart'
    show Content;
import 'package:angular2/src/core/compiler/shadow_dom_emulation/light_dom.dart'
    show LightDom;
import 'package:angular2/src/core/compiler/view.dart' show View;
import 'package:angular2/src/core/compiler/view_container.dart'
    show ViewContainer;

@proxy
class FakeView implements View {
  var contentTags;
  var viewContainers;
  FakeView([containers = null]) {
    this.contentTags = [];
    this.viewContainers = [];
    if (isPresent(containers)) {
      ListWrapper.forEach(containers, (c) {
        if (c is FakeContentTag) {
          ListWrapper.push(this.contentTags, c);
        } else {
          ListWrapper.push(this.contentTags, null);
        }
        if (c is FakeViewContainer) {
          ListWrapper.push(this.viewContainers, c);
        } else {
          ListWrapper.push(this.viewContainers, null);
        }
      });
    }
  }
  noSuchMethod(i) {
    super.noSuchMethod(i);
  }
}
@proxy
class FakeViewContainer implements ViewContainer {
  var templateElement;
  var _nodes;
  var _contentTagContainers;
  FakeViewContainer(templateEl, [nodes = null, views = null]) {
    this.templateElement = templateEl;
    this._nodes = nodes;
    this._contentTagContainers = views;
  }
  nodes() {
    return this._nodes;
  }
  contentTagContainers() {
    return this._contentTagContainers;
  }
  noSuchMethod(i) {
    super.noSuchMethod(i);
  }
}
@proxy
class FakeContentTag implements Content {
  var select;
  var _nodes;
  var contentStartElement;
  FakeContentTag(contentEl, [select = '', nodes = null]) {
    this.contentStartElement = contentEl;
    this.select = select;
    this._nodes = nodes;
  }
  insert(nodes) {
    this._nodes = ListWrapper.clone(nodes);
  }
  nodes() {
    return this._nodes;
  }
  noSuchMethod(i) {
    super.noSuchMethod(i);
  }
}
main() {
  describe('LightDom', () {
    var lightDomView;
    beforeEach(() {
      lightDomView = new FakeView();
    });
    describe("contentTags", () {
      it("should collect content tags from element injectors", () {
        var tag = new FakeContentTag(el('<script></script>'));
        var shadowDomView = new FakeView([tag]);
        var lightDom =
            new LightDom(lightDomView, shadowDomView, el("<div></div>"));
        expect(lightDom.contentTags()).toEqual([tag]);
      });
      it("should collect content tags from ViewContainers", () {
        var tag = new FakeContentTag(el('<script></script>'));
        var vc = new FakeViewContainer(null, null, [new FakeView([tag])]);
        var shadowDomView = new FakeView([vc]);
        var lightDom =
            new LightDom(lightDomView, shadowDomView, el("<div></div>"));
        expect(lightDom.contentTags()).toEqual([tag]);
      });
    });
    describe("expandedDomNodes", () {
      it("should contain root nodes", () {
        var lightDomEl = el("<div><a></a></div>");
        var lightDom = new LightDom(lightDomView, new FakeView(), lightDomEl);
        expect(toHtml(lightDom.expandedDomNodes())).toEqual(["<a></a>"]);
      });
      it("should include view container nodes", () {
        var lightDomEl = el("<div><template></template></div>");
        var lightDom = new LightDom(new FakeView([
          new FakeViewContainer(DOM.firstChild(lightDomEl), [el('<a></a>')])
        ]), null, lightDomEl);
        expect(toHtml(lightDom.expandedDomNodes())).toEqual(["<a></a>"]);
      });
      it("should include content nodes", () {
        var lightDomEl = el("<div><content></content></div>");
        var lightDom = new LightDom(new FakeView([
          new FakeContentTag(DOM.firstChild(lightDomEl), '', [el('<a></a>')])
        ]), null, lightDomEl);
        expect(toHtml(lightDom.expandedDomNodes())).toEqual(["<a></a>"]);
      });
      it("should work when the element injector array contains nulls", () {
        var lightDomEl = el("<div><a></a></div>");
        var lightDomView = new FakeView();
        var lightDom = new LightDom(lightDomView, new FakeView(), lightDomEl);
        expect(toHtml(lightDom.expandedDomNodes())).toEqual(["<a></a>"]);
      });
    });
    describe("redistribute", () {
      it("should redistribute nodes between content tags with select property set",
          () {
        var contentA = new FakeContentTag(null, "a");
        var contentB = new FakeContentTag(null, "b");
        var lightDomEl = el("<div><a>1</a><b>2</b><a>3</a></div>");
        var lightDom = new LightDom(
            lightDomView, new FakeView([contentA, contentB]), lightDomEl);
        lightDom.redistribute();
        expect(toHtml(contentA.nodes())).toEqual(["<a>1</a>", "<a>3</a>"]);
        expect(toHtml(contentB.nodes())).toEqual(["<b>2</b>"]);
      });
      it("should support wildcard content tags", () {
        var wildcard = new FakeContentTag(null, '');
        var contentB = new FakeContentTag(null, "b");
        var lightDomEl = el("<div><a>1</a><b>2</b><a>3</a></div>");
        var lightDom = new LightDom(
            lightDomView, new FakeView([wildcard, contentB]), lightDomEl);
        lightDom.redistribute();
        expect(toHtml(wildcard.nodes()))
            .toEqual(["<a>1</a>", "<b>2</b>", "<a>3</a>"]);
        expect(toHtml(contentB.nodes())).toEqual([]);
      });
    });
  });
}
toHtml(nodes) {
  if (isBlank(nodes)) return [];
  return ListWrapper.map(nodes, DOM.getOuterHTML);
}
